/*! \file  SphericalOptimalTransportationMesh.h
*   \brief Spherical Optimal Transport DynamicMesh
*   \author David Gu
*   \date   documented on 11/03/2020
*
*/
#ifndef  _SPHERICAL_OPTIMAL_TRANSPORT_DYNAMIC_MESH_H_
#define  _SPHERICAL_OPTIMAL_TRANSPORT_DYNAMIC_MESH_H_

#include <map>
#include <vector>
#include <stack>
#include <Eigen/Eigen>

#include "Mesh/Vertex.h"
#include "Mesh/Edge.h"
#include "Mesh/Face.h"
#include "Mesh/HalfEdge.h"
#include "Mesh/DynamicMesh.h"
#include "Mesh/Boundary.h"
#include "Mesh/Iterators.h"
#include "Parser/parser.h"
#include "SphericalPowerDiagramMesh.h"

#define T_SPD_MESH CSphericalPowerDiagramDynamicMesh<V, E, F, H>

namespace MeshLib
{
	/*! \brief COTFace class
	*
	*	Face class for optimal mass transport
	*   Trait: face normal
	*/
	class COMTFace : public CPowerFace
	{
	protected:
	public:
	};

	/*! \brief COTVertex class
	*
	*   Vertex class for optimal mass transport
	*   Trait : vertex rgb color
	*/
	class COMTVertex : public  CPowerVertex
	{
	protected:
		/*! vertex target area */
		double   m_target_area;
		/*! power cell area */
		double   m_dual_area;
		/*! vertex index */
		int		 m_index;
		/*! update direction */
		double	 m_update_direction;
		/*! backup phi */
		CPoint	 m_backup_phi;

	public:
		/*! CViewerVertex Constructor */
		COMTVertex() { m_index = 0; m_dual_area = 0; m_target_area = 0; }
		/*! vertex target area */
		double& target_area() { return m_target_area; };
		/*! vertex area */
		double& area() { return m_target_area; };
		/*! vertex dual area */
		double& dual_area() { return m_dual_area; };
		/*! vertex index*/
		int& index() { return m_index; };
		/*! update direction */
		double& update_direction() { return m_update_direction; };
		/*! backup height */
		CPoint& backup_phi() { return m_backup_phi; };
		/*! read from string */
		void _from_string();
		/*! write to string */
		void _to_string();
	};
	// read vertex rgb, uv from vertex string 
	inline void COMTVertex::_from_string()
	{
		CParser parser(m_string);

		for (std::list<CToken*>::iterator iter = parser.tokens().begin(); iter != parser.tokens().end(); ++iter)
		{
			CToken* token = *iter;
			if (token->m_key == "uv")
			{
				token->m_value >> m_uv;
			}
			else if (token->m_key == "rgb")
			{
				token->m_value >> m_rgb;
			}
			else if (token->m_key == "normal")
			{
				token->m_value >> m_normal_map;
			}
			else if (token->m_key == "u")
			{
				token->m_value >> m_phi;
			}

		}
	};

	//converting vertex uv trait to string

	inline void COMTVertex::_to_string()
	{
		CParser parser(m_string);
		parser._removeToken("u");
		parser._removeToken("c");
		parser._toString(m_string);
		std::stringstream iss;
		iss << "c=(" << m_cell_center[0] << " " << m_cell_center[1] << " " << m_cell_center[2] << ")";
		iss << " ";
		iss << "u=(" << m_phi[0] << " " << m_phi[1] << " " << m_phi[2] << ")";
		if (m_string.length() > 0)
		{
			m_string += " ";
		}
		m_string += iss.str();
	}

	/*! \brief COTEdge class
	*
	*   Edge class for optimal transport
	*   Trait : Edge sharp
	*/
	class COMTEdge : public  CPowerEdge
	{
	protected:
		/*! edge weight */
		double m_weight;
	public:
		/*! edge weight for Hessian matrix */
		double& weight() { return m_weight; };
	};

	/*! \brief COMTHalfEdge class
	 *
	 *   HalfEdge class for optimal mass transport
	 */
	class COMTHalfEdge : public  CPowerHalfEdge
	{
	public:
	protected:
	};

	/*-------------------------------------------------------------------------------------------------------------------------------------

		Optimal Mass Transport Mesh

	--------------------------------------------------------------------------------------------------------------------------------------*/
	/*! \brief COMTMesh class
	*
	*	mesh class for optimal mass transport
	*
	*/
	template<typename V, typename E, typename F, typename H>
	class CSOTDynamicMesh : public CSphericalPowerDiagramDynamicMesh<V, E, F, H>
	{
	public:
		typedef typename T_SPD_MESH::MeshVertexIterator MeshVertexIterator;
        typedef typename T_SPD_MESH::MeshEdgeIterator MeshEdgeIterator;

		//newton's method for computing update direction
		virtual void _update_direction();
		//compute L2 error and relative error
		virtual double _error();

	protected:
		//compute the gradient
		virtual void _calculate_gradient();
		//compute the Hessian
		virtual void __calculate_hessian(Eigen::SparseMatrix<double>& m_hessian);
		//solve the Hessian system
		virtual int  __solve(Eigen::SparseMatrix<double>& m_hessian);

	};

	typedef MeshLib::CSOTDynamicMesh<COMTVertex, COMTEdge, COMTFace, COMTHalfEdge> CSOTMesh;

	/*! Compute L2 error and maximal relative error */
	template<typename V, typename E, typename F, typename H>
	double CSOTDynamicMesh<V, E, F, H>::_error()
	{
		double max_error = -1e+10;
		double total_error = 0;

		for (MeshVertexIterator viter(this); !viter.end(); viter++)
		{
			V* pv = *viter;

			double   da = fabs(pv->target_area() - pv->dual_area());
			double error = da / pv->target_area();
			if (error > max_error)
			{
				max_error = error;
			}
			total_error += da * da;
		}
		std::cout << "Max relative error is " << max_error << " Total L2 error is " << total_error << std::endl;
		return max_error;
	}

	/*! compute the gradient of the energy */
	template<typename V, typename E, typename F, typename H>
	void CSOTDynamicMesh<V, E, F, H>::_calculate_gradient()
	{
		for (MeshVertexIterator viter(this); !viter.end(); viter++)
		{
			V* pv = *viter;
			//pv->update_direction() = (pv->target_area() - pv->dual_area())*pv->phi().norm();
			pv->update_direction() = (pv->target_area() - pv->dual_area());
		}
	};
	
	/*! Compute Hessian Matrix */
	template<typename V, typename E, typename F, typename H>
	void CSOTDynamicMesh<V, E, F, H>::__calculate_hessian(Eigen::SparseMatrix<double>& m_hessian)
	{
		m_hessian.resize(this->numVertices(), this->numVertices());
		std::vector<Eigen::Triplet<double>> triplets;


		for (MeshEdgeIterator eiter(this); !eiter.end(); eiter++)
		{
			E* pe = *eiter;
			V* pv0 = this->edgeVertex1(pe);
			V* pv1 = this->edgeVertex2(pe);
			int idx0 = pv0->index();
			int idx1 = pv1->index();
			double h = pe->weight();

			triplets.push_back({ idx0, idx1, h / 2 });
			triplets.push_back({ idx1, idx0, h / 2 });
			triplets.push_back({ idx0, idx0, -h / 2 });
			triplets.push_back({ idx1, idx1, -h / 2 });
		}
		m_hessian.setFromTriplets(triplets.begin(), triplets.end());
	}

	/*! solve linear system */
	template<typename V, typename E, typename F, typename H>
	int CSOTDynamicMesh<V, E, F, H>::__solve(Eigen::SparseMatrix<double>& m_hessian)
	{
		Eigen::VectorXd m_gradient;
		m_gradient.resize(this->numVertices());
		for (MeshVertexIterator viter(this); !viter.end(); viter++)
		{
			V* pv = *viter;
			double da = pv->target_area() - pv->dual_area();
			m_gradient[pv->index()] = pv->update_direction();
		}

		Eigen::SimplicialLDLT<Eigen::SparseMatrix<double>> solver;
		Eigen::VectorXd m_x;

		//hessian.coeffRef(0, 0) -= 1;
		solver.compute(m_hessian);
		m_x = solver.solve(m_gradient);
		// use conjugate gradient solver here
		//solver2.compute(hessian);
		//x = solver2.solve(gradient);
		m_x.array() -= m_x.mean();

		for (MeshVertexIterator viter(this); !viter.end(); viter++)
		{
			V* pv = *viter;
			pv->update_direction() = m_x[pv->index()];
		}
		return 0;
	}

	//use Newton's method to compute the update direction
	template<typename V, typename E, typename F, typename H>
	void CSOTDynamicMesh<V, E, F, H>::_update_direction()
	{
		_calculate_gradient();
		this->_edge_weight();
		Eigen::SparseMatrix<double> hessian;
		__calculate_hessian(hessian);
		__solve(hessian);
	}

	//damping algorithm for computing OT Map
	template<class COMTMesh>
	void _OT_damping(COMTMesh& mesh, double step )
	{
		//set the step length
		double step_length = step;
	
		//backup the current height
		for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
		{
			typename COMTMesh::CVertex* pv = *viter;
			pv->backup_phi() = pv->phi();
		}

		//make sure all the samples with heights are on the convex position
		while (true)
		{
			for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
			{
				typename COMTMesh::CVertex* pv = *viter;
				double d = step_length * pv->update_direction();
				d = exp(d);
				pv->phi() *= d;
			}
			//construct convex hull
			if (mesh._Lawson_edge_swap()) break;
			//roll back
			std::cout << "Roll Back" << std::endl;

			for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
			{
				typename COMTMesh::CVertex* pv = *viter;
				pv->phi() = pv->backup_phi();
			}
			step_length /= 2.0;
		}
		//make sure all power cells are non-empty
		while (true)
		{
			mesh._Legendre_transform();
			mesh._power_cell_area();

			bool legal = true;
			for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
			{
				typename COMTMesh::CVertex* pv = *viter;
				if (fabs(pv->dual_area()) < pv->target_area() * 1e-10) {
					legal = false;
					break;
				}
			}
			if (legal) break;
			//roll back
			std::cout << "Empty power cell. Roll Back" << std::endl;
			for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
			{
				typename COMTMesh::CVertex* pv = *viter;
				pv->phi() = pv->backup_phi();
			}
			step_length /= 2.0;
			for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
			{
				typename COMTMesh::CVertex* pv = *viter;
				double d = -step_length * pv->update_direction();
				d = exp(d);
				pv->phi() *= d;
			}
			mesh._Lawson_edge_swap();
		}
	};

	//Newton's method for computing OT map
	template<class COMTMesh>
	void OT_Newton(COMTMesh& mesh, const double step_length )
	{
		mesh._update_direction();
		_OT_damping<COMTMesh>(mesh, step_length );
		mesh._error();

		/*! debug */
		double epsilon = 1e-10;
		for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
		{
			typename COMTMesh::CVertex* pv = *viter;
			bool legal = false;
			for (typename COMTMesh::VertexEdgeIterator veiter(pv); !veiter.end(); veiter++)
			{
				typename COMTMesh::CEdge* pe = *veiter;
				if (fabs(pe->weight()) > epsilon)
				{
					legal = true;
					break;
				}
			}
			if (!legal)
			{
				std::cout << "Illegal Vertex " << pv->id() << std::endl;
				break;
			}
		}
	};

	//initalize for computing OT map
	template<class COMTMesh>
	void OT_Initialize(COMTMesh& mesh)
	{
		//set the vertex index
		int idx = 0;
		for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
		{
			typename COMTMesh::CVertex* pv = *viter;
			pv->index() = idx++;
		}

		//normalize the target area
		COperator<COMTMesh> pS(&mesh);
		pS._calculate_face_vertex_area();
	
		double total_area = 0;
		for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
		{
			typename COMTMesh::CVertex* pv = *viter;
			total_area += pv->target_area();
		}
		double sphere_area = 4 * PI;

		for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
		{
			typename COMTMesh::CVertex* pv = *viter;
			pv->target_area() *= (sphere_area / total_area);
		}
		//set the inital height funciton
		for (typename COMTMesh::MeshVertexIterator viter(&mesh); !viter.end(); viter++)
		{
			typename COMTMesh::CVertex* pv = *viter;
			//specially handle the inifinity vertex
			pv->phi() /= pv->phi().norm();
		}

		mesh._Lawson_edge_swap();
		mesh._Legendre_transform();
		mesh._power_cell_area();
	}
}

#undef T_SPD_MESH
#endif  //! SPHERICAL_OPTIMAL_TRANSPORT_DYNAMIC_MESH_H_
